from io import StringIO

import openpyxl
import pandas as pd


def convert_lcia_implementation_to_nis(lcia_implementation_file: str, lcia_file: str):
    def get_horizon(ind):
        return ""

    workbook = openpyxl.load_workbook(lcia_implementation_file, data_only=True)
    cfs = workbook["CFs"]
    units = workbook["units"]
    methods = set()
    categories = set()
    indicators = {}  # And its unit
    horizons = set()
    compartments = {}  # Include subcompartments
    for r in range(1, units.max_row):
        method = units.cell(row=r, column=1).value.strip()
        if "superseded" in method.lower() or "obsolete" in method.lower():
            continue
        if "recipe" not in method.lower() or "midpoint" not in method.lower():
            continue

        category = units.cell(row=r, column=2).value.strip()
        indicator = units.cell(row=r, column=3).value.strip()
        unit = units.cell(row=r, column=4).value.strip()

        methods.add(method)
        categories.add(category)
        indicators[indicator] = unit

    print("METHODS")
    print(methods)
    print("INDICATORS")
    print(indicators.keys())
    print("CATEGORIES")
    print(categories)
    _ = []  # Output
    for r in range(1, cfs.max_row):
        method = cfs.cell(row=r, column=1).value.strip()
        if method not in methods:
            continue
        category = cfs.cell(row=r, column=2).value.strip()
        indicator = cfs.cell(row=r, column=3).value.strip()
        # Guess horizon
        h = get_horizon(indicator)
        name = cfs.cell(row=r, column=4).value.strip()
        compartment = cfs.cell(row=r, column=5).value.strip()
        subcompartment = cfs.cell(row=r, column=6).value.strip()
        v = cfs.cell(row=r, column=7).value
        ind_unit = indicators[indicator]

        _.append((method, indicator, ind_unit, name, "", h, v, compartment, subcompartment))

    df = pd.DataFrame(_, columns=["LCIAMethod", "LCIAIndicator", "LCIAIndicatorUnit",
                                  "Interface", "InterfaceUnit",
                                  "LCIAHorizon", "LCIACoefficient",
                                  "Compartment", "Subcompartment"])

    s = StringIO()
    s.write("# To regenerate this file, execute action 'lcia_implementation_to_nis'\n")
    df.to_csv(s, index=False)
    with open(lcia_file, "wt") as f:
        f.write(s.getvalue())


if __name__ == '__main__':
    convert_lcia_implementation_to_nis("/home/rnebot/GoogleDrive/AA_SENTINEL/LCIA_implementation_3.7.1.xlsx", "/home/rnebot/Downloads/test.csv")



"""

Read a Sentinel model definition and outputs, to obtain one or more DiGraphs which can be used to
obtain either a MuSIASEM or LCA structure

"""
from nexinfosys.command_generators.parser_ast_evaluators import get_nis_name
from nexinfosys.common.helper import PartialRetrievalDictionary
from enbios.common.helper import list_to_dataframe, get_scenario_name
from enbios.input import Simulation
from enbios.model import SimStructuralProcessorAttributes
from friendly_data.dpkg import read_pkg
from friendly_data.converters import to_df


class SentinelSimulation(Simulation):
    def __init__(self, sentinel_index_path):
        self._sentinel_index_path = sentinel_index_path

    def read_raw(self):
        """ Read variables and datasets """
        pkg = read_pkg(self._sentinel_index_path)
        lst = []
        for r in pkg.resources:
            df = to_df(r)
            lst.append(df)

    def read(self, filter_model: str):
        """
        Reads Sentinel package completely

        :param filter_model: Read only the specified model
        :return: A tuple with assorted elements, from lists of possible values for dimensions to a registry of Processors
        """
        if filter_model is None:  # Mandatory
            return

        scenarios = set()
        regions = set()
        times = set()
        techs = set()
        carriers = set()
        units = set()
        col_types = set()  # Interface types
        ctc = set()  # Country - Tech - Carrier
        # print(f"SENTINEL INDEX: {self._sentinel_index_path}")
        pkg = read_pkg(self._sentinel_index_path)
        prd = PartialRetrievalDictionary()
        for res in pkg.resources:
            df = to_df(res)
            # print(f"INDEX: {df.index.names}; COLUMNS: {df.columns}")
            col_types.update(df.columns)
            region_idx = df.index.names.index("region") if "region" in df.index.names else -1
            carrier_name = "carrier" if "carrier" in df.index.names else "carriers" if "carriers" in df.index.names else None
            carrier_idx = df.index.names.index(carrier_name) if carrier_name is not None else -1
            tech_idx = df.index.names.index("technology") if "technology" in df.index.names else -1
            scenario_idx = df.index.names.index("scenario") if "scenario" in df.index.names else -1
            time_idx = df.index.names.index("year") if "year" in df.index.names else -1
            time_idx = time_idx if time_idx > -1 else df.index.names.index("time") if "time" in df.index.names else -1
            unit_idx = df.index.names.index("unit") if "unit" in df.index.names else -1
            for idx, cols in df.iterrows():
                if not isinstance(idx, tuple):  # When it is not a MultiIndex, Pandas has "idx" to not be a Tuple; workaround: convert into a Tuple of a single element
                    idx = (idx,)
                region = idx[region_idx] if region_idx >= 0 else None
                carrier = idx[carrier_idx] if carrier_idx >= 0 else None
                tech = idx[tech_idx] if tech_idx >= 0 else None
                scenario = idx[scenario_idx] if scenario_idx >= 0 else None
                time_ = str(idx[time_idx]) if time_idx >= 0 else None
                unit_ = idx[unit_idx] if unit_idx >= 0 else None
                if region:
                    regions.add(region)
                if carrier:
                    carriers.add(carrier)
                if scenario:
                    scenarios.add(scenario)
                if time_:
                    times.add(time_)
                if tech:
                    techs.add(tech)
                if tech and region:
                    ctc.add((tech, region, carrier))  # Carrier can be None
                if unit_:
                    units.add(unit_)
                if not region and not scenario and not time_:
                    region = scenario = time_ = "-"
                # -- Add COLS information --
                # (tech, region, scenario, time) - (cols)
                k = SimStructuralProcessorAttributes.partial_key(tech, region, carrier, scenario, time_)
                o = prd.get(k)
                if len(o) == 0:
                    pa = SimStructuralProcessorAttributes(tech, region, carrier, scenario, time_)
                    prd.put(k, pa)
                elif len(o) == 1:
                    pa = o[0]
                else:
                    raise Exception(f"Found {len(o)} occurrences of SimStructuralProcessorAttributes: {k}")
                pa.attrs.update({k: cols[k] for k in df.columns})  # Add variables from current pd.DataFrame
        return prd, scenarios, regions, times, techs, carriers, units, col_types, ctc

    def generate_template_data(self):
        """
        Produce a set of (tec, country) tuples and a set of (tec, -) tuples
        to help in producing "correspondence"
        :return:
        """
        prd, scenarios, regions, times, techs, carriers, units, col_types, ct = self.read("")

        cmds = []
        # Not NIS
        lst = [["technology", "calliope_technology_type"]]
        for t in techs:
            lst.append([t, ""])
        cmds.append(("Simulation Technology Types", list_to_dataframe(lst)))

        lst = [["region"]]
        for r in regions:
            lst.append([r])
        cmds.append(("Simulation Regions", list_to_dataframe(lst)))

        lst = [["carrier"]]
        for r in carriers:
            lst.append([r])
        cmds.append(("Simulation Carriers", list_to_dataframe(lst)))

        lst = [["period"]]
        for r in times:
            lst.append([r])
        cmds.append(("Simulation Times", list_to_dataframe(lst)))

        lst = [["unit"]]
        for r in units:
            lst.append([r])
        cmds.append(("Simulation Units", list_to_dataframe(lst)))

        lst = [["name", "region", "match_target_type", "match_target", "weight", "match_conditions"]]
        for t in techs:
            lst.append([t, "", "musiasem", "<musiasem parent>", 1, ""])
        for t in techs:
            lst.append([t, "", "lca", "<spold>", 1, ""])
        for t in ct:
            lst.append([t[0], t[1], "musiasem", "<musiasem parent>", 1, ""])
        for t in ct:
            lst.append([t[0], t[1], "lca", "<spold>", 1, ""])
        cmds.append(("Enbios Correspondence", list_to_dataframe(lst)))

        # InterfaceTypes
        lst = [["InterfaceTypeHierarchy", "InterfaceType", "Sphere", "RoegenType", "ParentInterfaceType", "Formula",
                "Description", "Unit", "OppositeSubsystemType", "Attributes"]]
        for col_type in col_types:
            lst.append(["sentinel", col_type, "Technosphere", "Flow", "", "",
                        "<description>", "<unit>", "<opposite>", ""])
        cmds.append(("InterfaceTypes", list_to_dataframe(lst)))

        # Georeferences
        geo_references = []
        for r in regions:
            # Georeferences (regions)
            print(r)

        # ProblemStatement
        lst = [["Scenario", "Parameter", "Value", "Description"]]
        for s in scenarios:
            # Scenario, Parameter, Value, Description
            scenario = get_scenario_name("s", s)
            observer = get_scenario_name("o", s)
            lst.append([scenario, "NISSolverObserversPriority", observer, f"Scenario {scenario}, observer {observer}"])
        cmds.append(("ProblemStatement", list_to_dataframe(lst)))

        # Processors, Interfaces and "Cloners"
        processors = [["ProcessorGroup", "Processor", "ParentProcessor", "SubsystemType", "System", "FunctionalOrStructural",
                "Accounted", "Stock", "Description", "GeolocationRef", "GeolocationCode", "GeolocationLatLong",
                "Attributes", "@tech_name", "@region"]]
        interfaces = [
            ["Processor", "InterfaceType", "Interface", "Sphere", "RoegenType", "Orientation", "OppositeSubsystemType",
             "GeolocationRef", "GeolocationCode", "InterfaceAttributes", "Value", "Unit", "RelativeTo", "Uncertainty",
             "Assessment", "PedigreeMatrix", "Pedigree", "Time", "Source", "NumberAttributes", "Comments"]]
        cloners = [["InvokingProcessor", "RequestedProcessor", "ScalingType", "InvokingInterface",
                    "RequestedInterface", "Scale"]]
        country_level_procs_already_added = set()
        tech_procs_already_added = set()
        system_per_region = False
        for t in ct:
            o = prd.get(SimStructuralProcessorAttributes.partial_key(t[0], t[1]))
            if len(o) > 0:
                parent = ""  # TODO
                if t[1]:  # Region
                    name = f"ES_{t[1]}"
                    parent = name
                    system = t[1] if system_per_region else ""
                    if name not in country_level_procs_already_added:
                        processors.append(["", name, "", "", system, "Functional", "Yes", "",
                                           f"Country level processor, {t[1]}", "", "", "", "", t[1], t[1]])
                        # Clone (clone dendrogram inside "region" processor)
                        # TODO InvokingInterface, RequestedInterface, Scale are mandatory; specify something here
                        cloners.append((name, "EnergySector", "Clone", "", "", ""))
                        country_level_procs_already_added.add(name)
                else:
                    system = ""
                # BareProcessors
                o1 = o[0]  # prd.get(ProcessorAttributes.partial_key(t[0], t[1]), full_key=True)[0]
                original_name = t[0]
                name = get_nis_name(t[0])
                full_name = f"{parent}{'.' if parent!='' else ''}{name}"
                if name not in tech_procs_already_added:
                    add = True
                else:
                    add = False

                # Technology (structural) processor
                description = f'{o1.attrs.get("description", name)} at {t[1]}'
                processors.append(["", name, parent, "", system, "Structural", "Yes", "",
                                   description, "", "", "", "", original_name, t[1]])
                # Add observations (Interfaces)
                for p in o:
                    observer = ""
                    time_ = "Year"
                    # name = ""
                    region = ""
                    _ = p.attrs.copy()
                    if "scenario" in p.attrs:
                        observer = get_scenario_name("o", p.attrs["scenario"])
                        del _["scenario"]
                    if "technology" in p.attrs:
                        name_ = p.attrs["technology"]
                        del _["technology"]
                    if "region" in p.attrs:
                        region = p.attrs["region"]
                        del _["region"]
                    if "year" in p.attrs or "time" in p.attrs:
                        if "year" in p.attrs:
                            time_ = p.attrs["year"]
                            del _["year"]
                        else:
                            time_ = p.attrs["time"]
                            del _["time"]
                    for k, v in _.items():
                        interfaces.append([full_name, k, "", "Technosphere", "Flow", "<orientation>", "", "", "", "", v,
                                           "", "<relative_to>", "", "", "", "", time_, observer, "", ""])

        cmds.append(("BareProcessors", list_to_dataframe(processors)))
        cmds.append(("ProcessorScalings", list_to_dataframe(cloners)))
        cmds.append(("Interfaces", list_to_dataframe(interfaces)))

        return cmds

    def generate_nis_models(self, split_by_region=False, split_by_scenario=True, split_by_period=True):
        """
        Produce a set of (tec, country) processors with interface values,
        separated by scenario (NIS observer) and year (NIS time)
        :return:
        """

        # Yield NIS commands and a descriptor of which region


dp = "/home/rnebot/Downloads/borrame/calliope-output"





class SimStructuralProcessorAttributes:
    """
    A class to store Processor attributes while a Sentinel datapackage is read
    and values are stored in a PartialRetrievalDictionary
    Attributes can be stored directly
    """
    def __init__(self, technology=None, region=None, carrier=None, scenario=None, time_=None):
        self.attrs = {}
        if technology:
            self.attrs["technology"] = technology
        if region:
            self.attrs["region"] = region
        if carrier:
            self.attrs["carrier"] = carrier
        if scenario:
            self.attrs["scenario"] = scenario
        if time_:
            self.attrs["time"] = time_

    @staticmethod
    def partial_key(technology=None, region=None, carrier=None, scenario=None, time=None):
        d = {}
        if technology:
            d["_t"] = technology
        if region:
            d["_g"] = region
        if carrier:
            d["_c"] = carrier
        if scenario:
            d["_s"] = scenario
        if time:
            d["_d"] = time
        return d

    def key(self):
        return self.partial_key(self.attrs.get("technology"),
                                self.attrs.get("region"),
                                self.attrs.get("carrier"),
                                self.attrs.get("scenario"),
                                self.attrs.get("time"))




# /usr/bin/env python
import locale
import os
import fire
import webbrowser

from nexinfosys.bin.cli_script import set_log_level_from_cli_param
from enbios.input.data_preparation.lci_to_nis import SpoldToNIS
from enbios.input.data_preparation.lcia_implementation_to_nis import convert_lcia_implementation_to_nis
from enbios.input.data_preparation.recipe_to_nis import convert_recipe_to_nis
from enbios.input.data_preparation.sentinel_to_nis_prep import sentinel_to_prep_file
from enbios.processing import read_parse_configuration
from enbios.processing.main import Enviro
from enbios.visualize import create_dashboard_app

"""
DEVELOPMENT CLI EXECUTION:
 
python -m enbios.bin.script
  <it will show the automatically generated help>

CLI EXECUTION ("enbios" because it is defined with that name in setup.py):
 
enbios (same as previous)
  <it will show the -automatically generated- help>
  
"""


class Enbios:
    @staticmethod
    def recipe_to_csv(recipe_file: str, lcia_file: str, log: str = None):  # TODO DEPRECATED, use "lcia_implementation_to_csv"
        """
        Convert an XLSX file with ReCiPe2016 to a CSV file ready to be declared in a DatasetDef which can be imported
        in a "LCIAMethod" command

        Example:
           enbios recipe_to_csv /home/rnebot/GoogleDrive/AA_SENTINEL/workspace_20210331/ReCiPe2016_CFs_v1.1_20180117.xlsx /home/rnebot/GoogleDrive/AA_SENTINEL/case_studies/library/recipe2016_2.csv


        :param recipe_file: The full path of the input ReCiPe2016 XLSX file
        :param lcia_file: The full path of the output CSV file
        :param log: Set log level to one of: Error (E, Err), Debug (D), Warning (W, Warn), Info (I), Off, Critical (Fatal)
        :return:
        """
        set_log_level_from_cli_param(log)
        convert_recipe_to_nis(recipe_file, lcia_file)

    @staticmethod
    def lcia_implementation_to_csv(lcia_implementation_file: str, lcia_file: str, log: str = None):
        """
        Convert an XLSX file with LCIA implementation (from Ecoinvent) to a CSV file ready to be declared in a
        DatasetDef which can later be imported in a "LCIAMethod" command

        Example:
           enbios lcia_implementation_to_csv /home/rnebot/GoogleDrive/AA_SENTINEL/LCIA_implementation_3.7.1.xlsx /home/rnebot/GoogleDrive/AA_SENTINEL/case_studies/library/lcia_implementation_nis.csv


        :param lcia_implementation_file: The full path of the input LCIA implementation XLSX file
        :param lcia_file: The full path of the output CSV file
        :param log: Set log level to one of: Error (E, Err), Debug (D), Warning (W, Warn), Info (I), Off, Critical (Fatal)
        :return:
        """
        set_log_level_from_cli_param(log)
        convert_lcia_implementation_to_nis(lcia_implementation_file, lcia_file)

    # def lci_to_interface_types(self, lci_file_path: str, csv_file: str):
    #     """
    #     Extract the interface types from an LCI file
    #
    #     :param lci_file_path: Path of a single .spold file
    #     :param csv_file: Output CSV with a listing of distinct interface types (potentially importable in InterfaceTypes)
    #     :return:
    #     """
    #     lci_to_interfaces_csv(lci_file_path, csv_file)

    @staticmethod
    def lci_to_nis(spold_files_folder: str, correspondence_path: str,
                   nis_base_url: str, nis_structurals_base_path: str, log: str = None):
        """
        Scan correspondence file AND a NIS file to extract a NIS file with the commands:
           - InterfaceTypes
           - BareProcessors
           - Interfaces
        The correspondence file should be a CSV file with the following header:
           name,match_target_type,match_target,weight,match_conditions
        Example:
           enbios lci_to_nis /home/rnebot/GoogleDrive/AA_SENTINEL/enviro_tmp/ /home/rnebot/GoogleDrive/AA_SENTINEL/enviro/sim_correspondence_example.csv https://docs.google.com/spreadsheets/d/15NNoP8VjC2jlhktT0A8Y0ljqOoTzgar8l42E5-IRD90/edit?usp=sharing /home/rnebot/Downloads/out.xlsx
           enbios lci_to_nis /home/rnebot/GoogleDrive/AA_SENTINEL/enviro_tmp/ "" https://docs.google.com/spreadsheets/d/1nYzphq1XYrezquW3yj7UiJ8sNJ8RzJ8gqErg3oVmw2Q/edit?usp=sharing /home/rnebot/Downloads/lci_to_nis_output.xlsx

        Example lines in this file:
           wind_onshore_competing,lca,2fdb6da1-8281-453d-8a7c-0184dc3586c4_66c93e71-f32b-4591-901c-55395db5c132.spold,,
           wind_onshore_competing,musiasem,Energy_system.Electricity_supply.Electricity_generation.Electricity_renewables.Electricity_wind.Electricity_wind_onshore,,


        :param spold_files_folder: Local folder where .spold files are stored
        :param correspondence_path: Correspondence file
        :param nis_base_url: URL of a NIS file that will be used as Base for the assembly of the model
        :param nis_structurals_base_path: File name of the output NIS that will contain the structural processors
        :param log: Set log level to one of: Error (E, Err), Debug (D), Warning (W, Warn), Info (I), Off, Critical (Fatal)
        :return:
        """
        set_log_level_from_cli_param(log)
        s2n = SpoldToNIS()
        s2n.spold2nis("generic_energy_production", spold_files_folder, correspondence_path, nis_base_url, nis_structurals_base_path)

    @staticmethod
    def sentinel_to_nis_preparatory(sentinel_data_package_json_path: str, nis_file: str, log: str = None):
        """
        Read a Sentinel Data Package and elaborate an XLSX file with NIS commands and complementary information
        describing what is inside the Sentinel Package (enumeration of Scenarios, Regions, etc.) and a template of
        CORRESPONDENCE file as one of the Worksheets.

        :param sentinel_data_package_json_path: The full path of the JSON file describing a Sentinel data package
        :param nis_file: Full path of the output XLSX NIS formatted file
        :param log: Set log level to one of: Error (E, Err), Debug (D), Warning (W, Warn), Info (I), Off, Critical (Fatal)
        :return:
        """
        set_log_level_from_cli_param(log)
        sentinel_to_prep_file(sentinel_data_package_json_path, nis_file)

    @staticmethod
    def enviro(cfg_file_path,
               just_one_fragment: bool = False,
               generate_nis_base_file: bool = False,
               generate_full_fragment_files: bool = False,
               generate_interface_results: bool = False,
               generate_indicators: bool = False,
               keep_min_fragment_files: bool = True,
               max_lci_interfaces: int = 0,
               n_cpus: int = 1,
               log: str = None,
               just_prepare_base: bool = False):
        """
        The main function of the package, reads the contents of the configuration file, which can be either a JSON or YAML
        file. Following an example of YAML file contents:
            nis_file_location: "https://docs.google.com/spreadsheets/d/12AlJ0tdu2b-cfalNzLqFYfiC-hdDlIv1M1pTE3AfSWY/edit#gid=1986791705"
            correspondence_files_path: ""
            simulation_type: sentinel
            simulation_files_path: ""
            output_directory: ""

        (you can copy and paste the example in an empty text file as reference, and ".yaml" as extension)

        :param cfg_file_path:
        :param just_one_fragment: True if only one of the fragments is to be computed, to test things
        :param generate_nis_base_file: True if the Base file should be generated (once) for testing purposes
        :param generate_full_fragment_files: True if the current fragment should be dumped into a NIS formatted XLSX file
        :param generate_interface_results: True if a CSV with values at interfaces should be produced, for each fragment
        :param generate_indicators: True if a CSV with indicators should be produced, for each fragment
        :param keep_min_fragment_files: If True, do not delete minimal NIS files submitted to NIS to compute indicators
        :param max_lci_interfaces: Max number of LCI interfaces to consider. 0 for all (default 0)
        :param n_cpus: Number of CPUs of the local computer used to perform the process (default 1, sequential)
        :param log: Set log level to one of: Error (E, Err), Debug (D), Warning (W, Warn), Info (I), Off, Critical (Fatal)
        :param just_prepare_base: True to only preparing Base file and exit
        :return:
        """
        # locale.setlocale(locale.LC_ALL, 'en_US')
        set_log_level_from_cli_param(log)
        t = Enviro()
        t.set_cfg_file_path(cfg_file_path)

        t.compute_indicators_from_base_and_simulation(just_one_fragment,
                                                      generate_nis_base_file,
                                                      generate_full_fragment_files,
                                                      generate_interface_results,
                                                      keep_min_fragment_files,
                                                      generate_indicators,
                                                      max_lci_interfaces,
                                                      n_cpus,
                                                      just_prepare_base)

    @staticmethod
    def visualize(cfg_file_path, open_browser=False, log: str = None):
        """
        Prepare a Dash server enabling basic visualization of results
            - scenario
            - region
            - year
            - processor
            - [carrier]
            - indicator

        :param cfg_file_path:
        :param open_browser: If True, open the link in the browser
        :param log:
        :return:
        """
        # Read data
        cfg_file_path = os.path.realpath(cfg_file_path)
        cfg = read_parse_configuration(cfg_file_path)
        base_dir = cfg["output_directory"]
        if not os.path.isabs(base_dir):
            base_dir = os.path.join(os.path.dirname(cfg_file_path), base_dir)

        app = create_dashboard_app(base_dir)

        # Launch Dash dashboard server
        host = "localhost"
        port = 8050
        address = f"http://{host}:{port}"
        print(f"Please, open {address}. Ctrl+C to Stop.")
        if open_browser:
            webbrowser.open(address)
        app.run_server(host=host, port=port, use_reloader=False)


def main():
    import platform
    os.environ["PAGER"] = "cat" if platform.system().lower() != "windows" else "-"
    fire.Fire(Enbios)


if __name__ == '__main__':
    main()



from abc import ABC


class Simulation(ABC):
    """ Abstract simulation reader (the implementation will depend on the type of simulator used) """
    def read(self, filter_model: str):
        """ Read"""
        pass

    def blocks(self):
        """ An iterator into the blocks (technologies) in the simulation """
        pass

    def block_information(self, block_name):
        """ Return information for block """
        pass





import math

from nexinfosys.model_services import State

materials = {
    "Aluminium",
    "Antimony",
    "Arsenic",
    "Baryte",
    "Beryllium",
    "Borates",
    "Cadmium",
    "Cerium",
    "Chromium",
    "Cobalt",
    "Copper",
    "Diatomite",
    "Dysprosium",
    "Europium",
    "Fluorspar",
    "Gadolinium",
    "Gallium",
    "Gold",
    "Gypsum",
    "IronOre",
    "KaolinClay",
    "Lanthanum",
    "Lead",
    "Lithium",
    "Magnesite",
    "Magnesium",
    "Manganese",
    "Molybdenum",
    "NaturalGraphite",
    "Neodymium",
    "Nickel",
    "Palladium",
    "Perlite",
    "Phosphorus",
    "Platinum",
    "Praseodymium",
    "Rhenium",
    "Rhodium",
    "Samarium",
    "Selenium",
    "SiliconMetal",
    "Silver",
    "Strontium",
    "Sulphur",
    "Talc",
    "Tantalum",
    "Tellurium",
    "Terbium",
    "Tin",
    "Titanium",
    "Tungsten",
    "Vanadium",
    "Yttrium",
    "Zinc",
    "Zirconium"
}


def supply_risk(state: State):
    sr = 0
    for i in materials:
        ri = state.get(i)
        if ri is not None:
            SRi = state.get(f"sr{i}")
            ci = state.get(f"c{i}")
            sr += ri*SRi/ci
    return sr


def recycling_rate(state: State):
    rr_num = 0
    rr_denom = 0
    for i in materials:
        ri = state.get(i)
        if ri is not None:
            RRi = state.get(f"rr{i}")
            rr_num += ri*RRi
            rr_denom += ri
    if rr_denom != 0.0:
        return rr_num / rr_denom
    else:
        return -1.0



#!/usr/bin/env python
from io import StringIO

import numpy as np
import openpyxl
import pandas as pd

"""
* Different names same substance: remove name. Use CIRpy if no reference name found
"""


def convert_recipe_to_nis(recipe_file, lcia_file):
    workbook = openpyxl.load_workbook(recipe_file, data_only=True)
    name_to_group = {"Global Warming": ("GWP", dict(name=1, alt_name= 2, formula=3, horizons=[4, 5, 6], row=4, unit_label_at=(2,), unit="kg")),
                     "Stratospheric ozone depletion": ("ODP", dict(name=1, alt_name=2, horizons=[3, 4, 5], row=5, unit_label_at=(2,), unit="kg")),
                     "Ionizing radiation": ("IRP", dict(name=1, alt_name=2, compartment=4, horizons=[5, 6, 7], row=4, unit_label_at=(2,), unit="kg")),
                     "Human damage ozone formation": ("HOFP", dict(name=2, compartment=4, horizons=[5, 6, 7], row=4, unit_label_at=(2,), unit="kg")),
                     "Particulate matter formation": ("PMFP", dict(name=2, horizons=[3, 4, 5], row=4, unit_label_at=(2,), unit="kg")),
                     "Ecosyste damage ozone formation": ("EOFP", dict(name=2, compartment=4, horizons=[5, 6, 7], row=4, unit_label_at=(2,), unit="kg")),
                     # "Terrestrial acidification": ("AP", dict(name=1, formula=1, horizons=[2, 3, 4], row=5, unit_label_at=(2,), unit="kg")),
                     # "Freshwater eutrophication": ("FEP", dict(name=1, compartment=2, formula=1, horizons=[3, 4, 5], row=5, unit_label_at=(2,), unit="kg")),
                     # "Marine eutrophication": ("MEP", dict(name=1, compartment=2, formula=1, horizons=[3, 3, 3], row=5, unit_label_at=(2,), unit="kg")),

                     # "Terrestrial ecotoxicity": ("ETPterrestrial", dict(name=2, compartment=4, formula=1, horizons=[5, 6, 7], row=4, unit_label_at=(2,), unit="kg")),
                     # "Freshwater ecotoxicity": ("ETPfw", dict(name=2, compartment=5, formula=1, horizons=[6, 7, 8], row=4, unit_label_at=(2,), unit="kg")),
                     # "Marine ecotoxicity": ("ETPmarine", dict(name=2, compartment=5, formula=1, horizons=[6, 7, 8], row=4, unit_label_at=(2,), unit="kg")),
                     # "Human carcinogenic toxicity": ("CFhuman_mid_carc", dict(name=2, compartment=5, formula=1, horizons=[6, 7, 8], row=4, unit_label_at=(2,), unit="kg")),
                     # "Human noncarcinogenic toxicity": ("CFhuman_mid_ncarc", dict(name=2, compartment=5, formula=1, horizons=[6, 7, 8], row=4, unit_label_at=(2,), unit="kg")),

                     ## The following are not easily treatable by LCIAMethods. Let keep them disabled ... (until a way to deal with them is decided)
                     # "Land transformation": ("LTx", dict(name=1, formula=1, horizons=[4, 5, 6], row=3, unit_label_at=None, unit="kg")),
                     # "Land occupation": ("LOcc", dict(name=1, formula=1, horizons=[4, 5, 6], row=5, unit_label_at=(2,), unit="kg")),
                     # "Water consumption": ("Wcons", dict(name=1, formula=1, horizons=[4, 5, 6], row=5, unit_label_at=(2,), unit="kg")),

                     # "Mineral resource scarcity": ("SOP", dict(name=1, formula=1, horizons=[3, 4, 5], row=4, unit_label_at=(2,), unit="kg")),
                     # "Fossil resource scarcity": ("FFP", dict(name=1, formula=1, horizons=[3, 4, 5], row=4, unit_label_at=(2,), unit="kg"))
                     }

    horizon_codes = ["I", "H", "E"]  # Individualist (20), Hierarchist (100), Egalitarian (>=1000)
    # TODO "Reference name" should be Ecospold name...
    reference_names = {}  # Reference names and list of alternative names.
    alt_names = {}  # Alternative to reference names

    _ = []  # Output
    for sheet_name in workbook.sheetnames:
        sheet = workbook[sheet_name]
        name = sheet.title.strip()
        if name not in name_to_group:
            continue
        print(name)
        tmp = name_to_group[name]
        ind = tmp[0]
        tmp = tmp[1]
        unit_label_at = tmp["unit_label_at"][0]
        unit = tmp["unit"]
        compartment_col = tmp.get("compartment")
        horizon_labels = [sheet.cell(row=unit_label_at, column=c).value for c in tmp["horizons"]]
        for r in range(tmp["row"], sheet.max_row):
            name = sheet.cell(row=r, column=tmp["name"]).value
            if name:
                name = name.strip()
            if not name:
                continue
            # TODO Find STANDARD name. Use CIRpy?
            name = name
            horizon_values = [sheet.cell(row=r, column=c).value for c in tmp["horizons"]]
            horizon_values = [h if h != "#N/A" else "" for h in horizon_values]
            compartment = sheet.cell(row=r, column=compartment_col).value if compartment_col else ""

            if horizon_values[0] is None or horizon_values[0] == "":
                continue
            for h, v in zip(horizon_codes, horizon_values):
                _.append(("ReCiPe2016", ind, name, unit, h, v, compartment))

    df = pd.DataFrame(_, columns=["LCIAMethod", "LCIAIndicator",
                                  "Interface", "InterfaceUnit",
                                  "LCIAHorizon", "LCIACoefficient",
                                  "Compartment"])

    s = StringIO()
    s.write("# To regenerate this file, execute module 'recipe_to_nis.py'\n")
    df.to_csv(s, index=False)
    with open(lcia_file, "wt") as f:
        f.write(s.getvalue())





# ONE TIME  -----------------------------------------------------------------------------
#
# in terminal
# # pip install --upgrade setuptools wheel twine
#
# Create account:
# PyPI test: https://test.pypi.org/account/register/
# or PyPI  : https://pypi.org/account/register/
#
# EACH TIME -----------------------------------------------------------------------------
#
# Modify version code in "setup.py" (this file)
#
# in python console
# Build (cd to directory where "setup.py" is)
# # import os
# # os.getcwd()
# python3 setup.py sdist bdist_wheel (is this code? It return error)
#
# Upload:
# PyPI test: twine upload --skip-existing --repository-url https://test.pypi.org/legacy/ dist/*
# or PyPI  : twine upload --skip-existing dist/*
#
# INSTALL   ------------------------------------------------------------------------------
#
# Local    : python setup.py install
# PyPI test: pip install --index-url https://test.pypi.org/simple/ --upgrade enbios
# PyPI     : pip install --upgrade enbios
# No PyPI  : pip install -e <local path where "setup.py" (this file) is located>
#
from os import path
from setuptools import setup
from pkg_resources import yield_lines
# from distutils.extension import Extension
# from Cython.Build import cythonize
# from Cython.Distutils import build_ext

"""
python3 setup.py sdist bdist_wheel
twine upload --skip-existing dist/*
"""

package_name = 'enbios'
version = '0.49'


def parse_requirements(strs):
    """Yield ``Requirement`` objects for each specification in `strs`

    `strs` must be a string, or a (possibly-nested) iterable thereof.
    """
    # create a steppable iterator, so we can handle \-continuations
    lines = iter(yield_lines(strs))

    ret = []
    for line in lines:
        # Drop comments -- a hash without a space may be in a URL.
        if ' #' in line:
            line = line[:line.find(' #')]
        # If there is a line continuation, drop it, and append the next line.
        if line.endswith('\\'):
            line = line[:-2].strip()
            try:
                line += next(lines)
            except StopIteration:
                return
        ret.append(line)

    return ret


with open('requirements.txt') as f:
    required = f.read().splitlines()

install_reqs = parse_requirements(required)
print(install_reqs)

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name=package_name,
    version=version,
    install_requires=install_reqs,
    packages=['enbios', 'enbios.bin', 'enbios.common', 'enbios.model',
              'enbios.input', 'enbios.input.data_preparation', 'enbios.input.simulators',
              'enbios.output',
              'enbios.visualize',
              'enbios.processing', 'enbios.processing.indicators'],
    include_package_data=True,
    url='https://github.com/ENVIRO-Module/enviro',
    license='MIT',
    author=['Rafael Nebot', 'Cristina Madrid'],
    author_email='rnebot@itccanarias.org',
    entry_points={'console_scripts': ['enbios=enbios.bin.script:main']},
    long_description=long_description,
    long_description_content_type='text/markdown',
    description='Indicators of environmental sustainability of energy systems using MuSIASEM and LCA methodologies'
)



import copy
import json
import os
import re
import tempfile
import urllib.request
from io import BytesIO
from typing import List, Union, BinaryIO
from zipfile import ZipFile

import numpy as np
import pandas as pd
from nexinfosys.command_generators import IType
from nexinfosys.common.helper import download_file
from nexinfosys.embedded_nis import NIS
from openpyxl import Workbook
from openpyxl.utils import get_column_letter
from openpyxl.utils.dataframe import dataframe_to_rows
from openpyxl.writer.excel import save_virtual_workbook

from enbios.processing import read_submit_solve_nis_file

JSON_INDENT = 4
ENSURE_ASCII = False


def _json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""
    from datetime import datetime

    if isinstance(obj, datetime):
        serial = obj.isoformat()
        return serial
    elif isinstance(obj, np.int64):
        return int(obj)
    elif getattr(obj, "Schema"):
        # Use "marshmallow"
        return getattr(obj, "Schema")().dump(obj)
    raise TypeError(f"Type {type(obj)} not serializable")


def generate_json(o):
    return json.dumps(o,
                      default=_json_serial,
                      sort_keys=True,
                      indent=JSON_INDENT,
                      ensure_ascii=ENSURE_ASCII,
                      separators=(',', ': ')
                      ) if o else None


def list_to_dataframe(lst: List) -> pd.DataFrame:
    return pd.DataFrame(data=lst[1:], columns=lst[0])


def get_scenario_name(prefix, s):
    def is_int(si):
        try:
            int(si)
            return True
        except ValueError:
            return False

    return f"{prefix}{s}" if is_int(s) else str(s)


def generate_workbook(cmds, generate_empty=False):
    # Convert list of pd.DataFrames to Excel workbook
    wb = Workbook(write_only=True)
    ws_count = 0
    for name, df in cmds:
        if df.shape[0] == 0:
            if not generate_empty:
                continue

        ws_count += 1
        ws = wb.create_sheet(name)
        widths = [0]*(df.shape[1]+1)  # A maximum of 100 columns
        max_columns = 0
        for r in dataframe_to_rows(df, index=False, header=True):
            if len(r) > max_columns:
                max_columns = len(r)
            for i in range(len(r)):
                width = int(len(str(r[i])) * 1.1)
                if width > widths[i]:
                    widths[i] = width

        for i, column_width in enumerate(widths):
            ws.column_dimensions[get_column_letter(i+1)].width = column_width

        for r in dataframe_to_rows(df, index=False, header=True):
            ws.append(r)

    if ws_count > 0:
        return save_virtual_workbook(wb)
    else:
        return None


def hash_array(f: Union[str, bytes]):
    import hashlib
    m = hashlib.md5()
    if isinstance(f, str):
        m.update(f.encode("utf-8"))
    else:
        m.update(f)
    return m.digest()


def set_zip_timestamp(in_zip: Union[os.PathLike, str, BinaryIO], timestamp=(2020, 1, 1, 0, 0, 0)) -> BinaryIO:
    """
    Modify the timestamp of all files, to stabilize the hash

    :param in_zip: Zip file whose files timestamp will be modified
    :param timestamp: Tuple with the timestamp to set for all the files
    :return A BytesIO with the resulting Zip
    """
    out_zip = BytesIO()
    with ZipFile(in_zip, mode="r") as zin:
        with ZipFile(out_zip, mode="w") as zout:
            for zinfo in zin.infolist():
                data = zin.read(zinfo.filename)
                zinfo_new = copy.copy(zinfo)
                zinfo_new.date_time = timestamp
                zout.writestr(zinfo_new, data)
    return out_zip


def get_valid_name(original_name):
    prefix = original_name[0] if original_name[0].isalpha() else "_"
    remainder = original_name[1:] if original_name[0].isalpha() else original_name
    return prefix + re.sub("[^0-9a-zA-Z_]+", "", remainder)


def get_file_url(fn):
    return "file:" + urllib.request.pathname2url(fn)


def prepare_base_state(base_url: str, solve: bool, directory: str = None):
    from nexinfosys import initialize_configuration
    initialize_configuration()  # Needed to make download and NIS later work properly
    tmp_io = download_file(base_url)
    bytes_io = set_zip_timestamp(tmp_io)
    hash = hash_array(bytes_io.getvalue())
    val_name = get_valid_name(base_url)
    if directory is None:
        directory = tempfile.gettempdir()
    hash_file = os.path.join(f"{directory}", f"base.hash.{val_name}")
    state_file = os.path.join(f"{directory}", f"base.state.{val_name}")
    if os.path.isfile(hash_file) and os.path.isfile(state_file):
        with open(hash_file, "rb") as f:
            cached_hash = f.read()
        update = cached_hash != hash
    else:
        update = True

    if update:
        temp_name = tempfile.NamedTemporaryFile(dir=directory, delete=False)
        temp_name = temp_name.name
        with open(temp_name, "wb") as f:
            f.write(bytes_io.getvalue())
        nis, issues = read_submit_solve_nis_file(get_file_url(temp_name), None, solve=solve)
        os.remove(temp_name)
        any_error = False
        for issue in issues:
            print(issue)
            if issue.itype == IType.ERROR:
                any_error = True
        # Write if there are no errors. IF there is any error generate an Exception
        if any_error:
            raise Exception(f'There were errors with the NIS base file {base_url}.')
        from nexinfosys.serialization import serialize_state
        serial_state = serialize_state(nis.get_state())
        nis.close_session()
        with open(hash_file, "wb") as f:
            f.write(hash)
        with open(state_file, "wb") as f:
            f.write(serial_state)
    else:
        with open(state_file, "rb") as f:
            serial_state = f.read()

    return serial_state



from typing import Dict, Tuple, List

from nexinfosys.command_generators import Issue
from nexinfosys.embedded_nis import NIS
from nexinfosys.model_services import State


def read_parse_configuration(file_name) -> Dict:
    """
    (read &) Parse the configuration file. See "example_config.yml"

    :param file_name: Name of the configuration file
    :return: If mandatory keys are present it returns a "dict" with every variable in the configuration file
    """
    def read_json_configuration(file_content) -> Dict:
        import json
        d = json.loads(file_content)
        return d

    def read_yaml_configuration(file_content) -> Dict:
        import yaml
        d = yaml.load(file_content, Loader=yaml.FullLoader)
        return d

    if isinstance(file_name, str):
        # Read into string
        with open(file_name, 'r') as file:
            contents = file.read()

        # Read depending on format
        if file_name.endswith("yaml") or file_name.endswith("yml"):
            cfg = read_yaml_configuration(contents)
        elif file_name.endswith("json"):
            cfg = read_json_configuration(contents)
    elif isinstance(file_name, dict):
        cfg = file_name

    # Check configuration
    mandatory_keys = ["nis_file_location",
                      "correspondence_files_path",
                      "simulation_type",
                      "simulation_files_path",
                      "output_directory"]
    any_not_found = False
    for k in mandatory_keys:
        if k not in cfg:
            print(f"{k} key not found in the configuration file {file_name}")
            any_not_found = True
    if any_not_found:
        raise Exception("Configuration file incomplete or incorrect. See log for missing keys")
    return cfg


def read_submit_solve_nis_file(nis_file_url: str, state: State, solve=False) -> Tuple[NIS, List[Issue]]:
    nis = NIS()
    nis.open_session(True, state)
    nis.reset_commands()
    # Load file and execute it
    nis.load_workbook(nis_file_url)
    if solve:
        issues = nis.submit_and_solve()
    else:
        issues = nis.submit()
    # tmp = nis.query_available_datasets()
    # print(tmp)
    # d = nis.get_results([(tmp[0]["type"], tmp[0]["name"])])
    return nis, issues


from enbios.common.helper import generate_workbook
from enbios.input.simulators.sentinel import SentinelSimulation


def sentinel_to_prep_file(sentinel_package_index, output):
    sim = SentinelSimulation(sentinel_package_index)
    cmds = sim.generate_template_data()
    s = generate_workbook(cmds, True)
    with open(output, "wb") as f:
        f.write(s)



"""

Forces in the energy system:

* Demand. Seems the driving force. But it must adapt to feasibility (of final configurations and
  the actionability of transitions)
* Economic. Investment and operation costs of final configuration and the transition to it.
* Technical. The scenario must meet all technical constraints. For instance, RE penetration is usually limited to
  a percentage of total demand.
* Social. jobs, decrease of energy-poverty, energy self-sufficiency

Generate a Calliope model.

Which information is available?
* impact of technologies. A balance of impacts may be performed


"""
